# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Module to build C/C++ extension with Poetry."""

import configparser
import os
import platform
from glob import glob
from os.path import join, relpath, splitext
from pathlib import Path

from setuptools.command.develop import develop
from setuptools.extension import Extension

# arch information comes from
# gcc -v
# gcc -dumpmachine

# info
#  https://stackoverflow.com/questions/6928110/how-may-i-override-the-compiler-gcc-flags-that-setup-py-uses-by-default

print(f"SYSTEM: {platform.system()}")
print(f"MACHINE: {platform.machine()}")
print(f"UNAME: {platform.uname()}")
print(f"RELEASE: {platform.release()}")
print(f"VERSION: {platform.version()}")
print(f"PROCESSOR: {platform.processor()}")

_platform = platform.system().lower()
_is_windows = 'windows' in _platform  # pylint: disable=C0103
_arch = 'x86-64'  # pylint: disable=C0103
_c_compiler = ""  # pylint: disable=C0103
if _is_windows:
    _lflags = f"-Wl,--subsystem,windows,--out-implib,libcode128_{_arch}.a"
    _c_compiler = "mingw32"  # pylint: disable=C0103
else:
    # if "linux" in _platform:
    _c_compiler = "unix"  # pylint: disable=C0103
    _lflags = "-Wl,-soname,libcode128.so"  # pylint: disable=C0103
    if 'arm' in platform.machine():
        _arch = "armv7"  # pylint: disable=C0103
    elif 'aarch64' in platform.machine():
        _arch = "armv8.8-a"  # pylint: disable=C0103

_march = f"-march={_arch}"  # pylint: disable=C0103
if 'darwin' in _platform:
    _march = ""  # pylint: disable=C0103  # clang compiler does not support -march=x86_64


# list of tuples (python_module, dict of C/C++ library build data ),
#   python_module is the extension that uses those C/C++ library info.
# the dict can be extended with all the kwargs needed by Extension to correctly build C/C++ sourcecode
_LIBS = [
    (
        'pycode128',
        {
            "libs": ['libs/code128'],
            # Building library part
            "lib_cflags": ["-O2", "-std=c99", "-Wall", "-fpic", "-Wextra", _march, "-DADD_EXPORTS"],
            "lib_lflags": ["-shared", _lflags],
            # Python extension part
            # always define PY_SSIZE_T_CLEAN , see https://docs.python.org/3/extending/extending.html
            "extension_define_macros": [("PY_SSIZE_T_CLEAN", None), ("ADD_EXPORTS", None)],
            "extension_language": "c99",
            "extension_extra_compile_args": [],
            "extension_extra_link_args": [],
            "extension_library": ['code128'],
        },
    )
]


class CustomDevelop(develop):
    """Custom install procedure.

    When declaring a ``build_extension.py`` poetry switches to setuptools during
    installation, i.e., it generates a temporary ``setup.py`` and then calls
    ``setup.py develop`` on it when you call ``poetry install``.
    Consequentially, we can hook into the develop command and customize the
    build to compile our source :) Note that this is only needed for the
    ``develop`` command, because the ``build`` command (``poetry build``)
    already includes ``build_clib``.

    This class then is the hook that will compile the source when we call
    ``poetry install``.

    """

    def run(self) -> None:  # type: ignore
        # build archives (.lib) these are declared in the `libraries` kwarg of
        # setup(). Extensions may depend on these, so we have to build the libs
        # them first.
        self.run_command("build_clib")
        super().run()


# custom_extension = Extension(
#     "pycode128.pycode128",
#     sources=["pycode128/pycode128.c"],
#     # define_macros=[("PY_SSIZE_T_CLEAN",)],
#     # we need to declare the extenal dependencies
#     include_dirs=["libs/code128"],
#     libraries=["code128"],  # see below
# )


def _add_compiler_to_setup_cfg(_compiler):
    conf = configparser.ConfigParser(inline_comment_prefixes="#")

    with open(Path(os.path.realpath(os.path.dirname(__file__))) / 'setup.cfg', mode='r', encoding="cp1252") as _fp:
        try:
            conf.read_file(_fp)
        except configparser.Error:
            pass

    try:
        conf.set('build_clib', 'compiler', _compiler)
    except configparser.NoSectionError:
        conf.add_section('build_clib')
        conf.set('build_clib', 'compiler', _compiler)

    try:
        conf.set('build_ext', 'compiler', _compiler)
    except configparser.NoSectionError:
        conf.add_section('build_ext')
        conf.set('build_ext', 'compiler', _compiler)

    with open('setup.cfg', mode='w', encoding="cp1252") as _fp:
        conf.write(_fp, space_around_delimiters=True)


# setuptools.extension requires unix separator
def _unix_form(file_path: str) -> str:
    return file_path.replace('\\', '/')


def build(setup_kwargs):
    """
    This is a callback for poetry used to hook in our extensions.
    """

    _add_compiler_to_setup_cfg(_c_compiler)

    setup_kwargs.update(
        {
            # declare archives (.lib) to build. These will be linked to
            # statically by extensions, cython, ...
            "libraries": [
                (
                    "code128",
                    {
                        "sources": [
                            _unix_form(path)
                            for _, _source_libs in _LIBS
                            for _source_folder in _source_libs['libs']
                            for root, _, _ in os.walk(os.sep.join([_source_folder]))
                            for path in glob(join(root, '*.c'))
                            if 'code128png' not in path
                        ],
                        # flags and dependencies of this library
                        # "include_dirs": ...
                        # "libraries": ...
                        "cflags": [
                            itm
                            for pylib, _source_libs in _LIBS
                            for itm in _source_libs['lib_cflags']
                            if pylib == 'pycode128'
                        ],
                        "lflags": [
                            itm
                            for pylib, _source_libs in _LIBS
                            for itm in _source_libs['lib_lflags']
                            if pylib == 'pycode128'
                        ],
                    },
                ),
            ],
            "ext_modules": [
                Extension(
                    splitext(relpath(_unix_form(path), start='.').replace(os.sep, '.'))[0],
                    sources=[_unix_form(path)],
                    define_macros=_source_libs['extension_define_macros'],
                    include_dirs=_source_libs['libs'],
                    libraries=_source_libs['extension_library'],
                    language=_source_libs["extension_language"],
                    extra_compile_args=_source_libs["extension_extra_compile_args"],
                    extra_link_args=_source_libs["extension_extra_link_args"],
                )
                for _py_lib, _source_libs in _LIBS
                for root, _, _ in os.walk(os.sep.join([_py_lib]))
                for path in glob(join(root, '*.c'))
            ],
            # hook into the build process to build our external sources before
            # we build and install the package.
            "cmdclass": {"develop": CustomDevelop},
        }
    )


#
# if __name__ == '__main__':
#     # _LIBS = [('py', 'libs/code128')]
#
#     for _py_lib, _source_libs in _LIBS:
#         for _source_folder in _source_libs['libs']:
#             for root, _, _ in os.walk(os.sep.join([_source_folder])):
#                 print(root)
#                 for path in glob(join(root, '*.c')):
#                     print(path)
#                     print(dirname(path))
#
#     _a = [_unix_form(path)
#           for _py_lib, _source_lib in _LIBS
#           for source_fold in _source_lib['libs']
#           for root, _, _ in os.walk(os.sep.join([source_fold]))
#           for path in glob(join(root, '*.c'))
#           if 'code128png' not in path]
#
#     for _py_lib, _ in _LIBS:
#         for root, _, _ in os.walk(os.sep.join([_py_lib])):
#             print(root)
#             for path in glob(join(root, '*.c')):
#                 print(path)
#                 print(dirname(path))
#                 print(splitext(relpath(path, start='.').replace(os.sep, '.'))[0])
#
#     ext = [Extension(splitext(relpath(_unix_form(path), start='.').replace(os.sep, '.'))[0],
#                      sources=[_unix_form(path)],
#                      define_macros=_source_libs['define_macros'],
#                      include_dirs=_source_libs['libs'])
#            for _py_lib, _source_libs in _LIBS
#            for root, _, _ in os.walk(os.sep.join([_py_lib]))
#            for path in glob(join(root, '*.c'))
#            ]
#     print(ext)
#
#



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Tests for `code128_image` module."""

from pathlib import Path

from PIL import Image, ImageChops

from pycode128.code128_image import Code128Image
from pycode128.pycode128 import PyCode128


def test_changing_image_bar_width():
    """Test image with different bar width."""
    _code128 = PyCode128('test')
    _code128.encode_raw()
    _image = Code128Image(_code128.encoded_data, bar_width=10)

    _image_expected = Image.open(Path.cwd().joinpath("tests").joinpath("image_width_10.png"))
    _diff = ImageChops.difference(_image.get_image(), _image_expected)
    assert not _diff.getbbox()


def test_changing_image_height():
    """Test image with different height."""
    _code128 = PyCode128('test')
    _code128.encode_raw()
    _image = Code128Image(_code128.encoded_data, image_height=100)

    _image_expected = Image.open(Path.cwd().joinpath("tests").joinpath("image_height_100.png"))
    _diff = ImageChops.difference(_image.get_image(), _image_expected)
    assert not _diff.getbbox()


def test_compare_image():
    """Compare image with existing one."""
    _code128 = PyCode128('test')
    _code128.encode_raw()
    _image_gen = Code128Image(_code128.encoded_data)
    _image_expected = Image.open(Path.cwd().joinpath("tests").joinpath("test_image.png"))
    _diff = ImageChops.difference(_image_gen.get_image(), _image_expected)
    assert not _diff.getbbox()


def test_compare_image_fnc3():
    """Compare image with existing one."""
    _code128 = PyCode128('[FNC3] $P\r')
    _code128.encode_raw()
    _image_gen = Code128Image(_code128.encoded_data)
    _image_expected = Image.open(Path.cwd().joinpath("tests").joinpath("fnc3_dollar_p.png"))
    _diff = ImageChops.difference(_image_gen.get_image(), _image_expected)
    assert not _diff.getbbox()



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Tests for `pycode128` cli tool."""

import os

from click.testing import CliRunner

from pycode128 import __version__
from pycode128.cli_tools import cli


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.pycode128)
    assert result.exit_code == 2
    assert 'Error: Missing argument' in result.output

    help_result = runner.invoke(cli.pycode128, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output

    version_result = runner.invoke(cli.pycode128, ['--version'])
    assert version_result.exit_code == 0
    assert __version__ in version_result.output


def test_command_line_argument():
    """Test the CLI arguments."""
    runner = CliRunner()

    barcode = runner.invoke(cli.pycode128, ['test'])
    assert barcode.exit_code == 0
    assert 'Input string: test' in barcode.output
    assert 'Barcode length: 99' in barcode.output

    out_file = runner.invoke(cli.pycode128, ['test', '-i', 'outfile.png'])
    assert out_file.exit_code == 0
    assert 'Saving image to file: outfile.png' in out_file.output
    os.unlink('outfile.png')


def test_command_line_mutually_exclusive():
    """Test the CLI arguments mutually exclusive."""
    runner = CliRunner()

    ret = runner.invoke(cli.pycode128, ['-c', '-r', 'testlabel'])
    assert 'Error: the following parameters are mutually exclusive:' in ret.output
    assert ret.exit_code == 2


def test_command_line_optional_arguments():
    """Test the CLI optional arguments."""
    runner = CliRunner()

    carriage_return = runner.invoke(cli.pycode128, ["-c", "TEST"])
    assert carriage_return.exit_code == 0
    assert 'Input string: TEST' in carriage_return.output
    assert 'Barcode length: 110' in carriage_return.output

    label_prog = runner.invoke(cli.pycode128, ["-r", "TEST"])
    assert label_prog.exit_code == 0
    assert 'Input string: [FNC3] $TEST' in label_prog.output
    assert 'Barcode length: 132' in label_prog.output



# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Console script for pycode128."""

import inspect
import types
from typing import Any, Callable, TypeVar, cast

from click import Choice, echo, version_option
from cloup import HelpFormatter, HelpTheme, Style, argument, command, option, option_group
from cloup.constraints import mutually_exclusive
from python_active_versions.utility import configure_logger

from pycode128 import __version__
from pycode128.code128_image import Code128Image
from pycode128.pycode128 import PyCode128  # pylint: disable=ungrouped-imports   # pylint: disable=no-name-in-module

F = TypeVar('F', bound=Callable[..., Any])


CONTEXT_SETTINGS = {"help_option_names": ['-h', '--help']}

formatter_settings = HelpFormatter.settings(
    theme=HelpTheme(
        invoked_command=Style(fg='bright_yellow'),
        heading=Style(fg='bright_white', bold=True),
        constraint=Style(fg='magenta'),
        col1=Style(fg='bright_yellow'),
    )
)


@command(context_settings=CONTEXT_SETTINGS, show_constraints=True, formatter_settings=formatter_settings)
@argument('data', help='input string to be converted as Code128 label')
@option("-i", "--image", "image_name", type=str, required=False, help="Output file name for generated image.")
@option_group(
    "Format Options",
    option("-c", "--carriage-return", "add_cr", is_flag=True, type=bool, required=False, help="Add ending CR."),
    option(
        "-r",
        "--reader-programming",
        "label_prog",
        is_flag=True,
        type=bool,
        required=False,
        help="Convert to Label Programming Barcode.",
    ),
    constraint=mutually_exclusive,
)
@option_group(
    "Generic Options",
    option(
        '-l',
        '--loglevel',
        'loglevel',
        type=Choice(["debug", "info", "warning", "error"], case_sensitive=False),
        default="info",
        show_default=True,
        help="set log level",
    ),
)
@version_option(__version__)
def pycode128(data: str, image_name: str, add_cr: bool, label_prog: bool, loglevel: str):
    """Code128 barcode generator library."""
    _fnc_name = cast(types.FrameType, inspect.currentframe()).f_code.co_name
    _complete_doc = inspect.getdoc(eval(_fnc_name))  # pylint: disable=eval-used  # nosec B307
    _doc = f"{_complete_doc}".split('\n')[0]  # use only doc first row
    _str = f"{_doc[:-1]} - v{__version__}"
    echo(f"{_str}")
    echo("=" * len(_str))

    configure_logger(loglevel)

    if add_cr:
        data = f"{data}\r"
    if label_prog:
        data = f"[FNC3] ${data}\r"
    _code128 = PyCode128(data)
    _code128.encode_raw()
    echo(f"Input string: {data}")
    echo(f"Barcode length: {_code128.length}")
    echo(f"Encoded data: {_code128.encoded_data}")

    if image_name:
        echo(f"Saving image to file: {image_name}")
        Code128Image(_code128.encoded_data).get_image().save(image_name)


if __name__ == "__main__":
    pycode128()  # pragma: no cover  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycode128', 'pycode128.cli_tools']

package_data = \
{'': ['*']}

install_requires = \
['click==8.1.3', 'cloup>=2.0.0,<3.0.0', 'pillow>=9.4.0,<10.0.0']

entry_points = \
{'console_scripts': ['pycode128 = pycode128.cli_tools.cli:pycode128']}

setup_kwargs = {
    'name': 'pycode128',
    'version': '2.2.0rc0',
    'description': 'Python extension for Code128 barcode generator library.',
    'long_description': "# Code128 library's python extension\n\n\n[![pypi](https://img.shields.io/pypi/v/pycode128.svg)](https://pypi.org/project/pycode128/)\n[![python](https://img.shields.io/pypi/pyversions/pycode128.svg)](https://pypi.org/project/pycode128/)\n[![Build Status](https://github.com/gpongelli/pycode128/actions/workflows/complete.yml/badge.svg)](https://github.com/gpongelli/pycode128/actions/workflows/complete.yml)\n[![codecov](https://codecov.io/gh/gpongelli/pycode128/branch/main/graphs/badge.svg)](https://codecov.io/github/gpongelli/pycode128)\n\n\n\nPython extension for Code128 barcode generator library\n\n\n* Documentation: <https://gpongelli.github.io/pycode128>\n* GitHub: <https://github.com/gpongelli/pycode128>\n* PyPI: <https://pypi.org/project/pycode128/>\n* Free software: MIT\n\n\n## Features\n\n* [Code128 library](https://github.com/fhunleth/code128) python wrapper, included as submodule\n  and build when creating python extension\n* Poetry managed project\n* C code checked via [check tool](https://libcheck.github.io/check/), compiled and run through pytst on all python version under Ubuntu OS\n* Multiple OS (Linux, Windows, MacOS) and Python compilation via [cibuildwhel](https://github.com/pypa/cibuildwheel)\n* [cibuildwheel manylinux_2_28 custom images](https://hub.docker.com/r/gpongelli/manylinux_python) with pre-built python from source,\n  to being able to compile check against correct python version\n* ~~[ARM-runner action](https://github.com/pguyot/arm-runner-action) custom [images](https://github.com/gpongelli/arm-runner-python/releases)\n  with pre-built python from source, to build RPI wheels~~\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [gpongelli/cookiecutter-pypackage](https://github.com/gpongelli/cookiecutter-pypackage) project template.\n",
    'author': 'Gabriele Pongelli',
    'author_email': 'gabriele.pongelli@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpongelli/pycode128',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4.0',
}
from build_extension import *
build(setup_kwargs)

setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2023 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Class to retrieve a Pillow Image from byte buffer."""

from itertools import repeat

from PIL import Image


class Code128Image:  # pylint: disable=too-few-public-methods
    """Class to convert byte buffer to PIL Image."""

    def __init__(self, byte_buffer: bytes, image_height: int = 200, bar_width: int = 5):
        """Initialization.

        Arguments:
            byte_buffer: byte buffer from PyCode128 C extension.
            image_height: height of final image
            bar_width: width of each single bar
        """
        self.__image_height = image_height
        self.__bar_width = bar_width

        # Once you get barcode_data, each byte corresponds to whether a vertical line should be drawn.
        # If the byte is 0xff, then draw a line. If the byte is 0x00, then don't.
        self.__bytes = byte_buffer.replace(b'\x00', b'\x01').replace(b'\xff', b'\x00').replace(b'\x01', b'\xff')

    def get_image(self) -> Image:
        """Retrieve PIL image.

        Returns:
            PIL Image object.
        """
        _cols = b''.join(map(lambda x: bytes(repeat(x, self.__bar_width)), self.__bytes))
        _image_width = len(_cols)
        _image_bytes = b''.join(repeat(_cols, self.__image_height))
        _image = Image.frombuffer(mode="L", size=(_image_width, self.__image_height), data=_image_bytes)
        return _image
